from django.contrib.auth.models import AbstractUser
from django.db import models
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.utils.text import slugify

from utils.model import BaseModel, FixCharField, DateTimeFormatField, MultipleChoiceField
from utils.util import cache


class Department(BaseModel):
    parent = FixCharField(max_length=32, default='12345678901234567890123456789012', verbose_name='上级目录')
    name = models.CharField(max_length=64, blank=True, null=True, verbose_name="部门名称")
    sort = models.SmallIntegerField(default=1, verbose_name='sort')
    is_active = models.BooleanField(default=True, verbose_name="是否启用")
    remark = models.CharField(max_length=128, blank=True, null=True, verbose_name="备注")

    class Meta:
        db_table = 'system_department'
        verbose_name = '部门表'
        verbose_name_plural = verbose_name


class User(AbstractUser, BaseModel):
    last_operation = DateTimeFormatField(blank=True, null=True, verbose_name="最后操作时间")
    nickname = models.CharField(max_length=32, blank=True, null=True, verbose_name="昵称")
    phone = models.CharField(max_length=11, blank=True, null=True, verbose_name="手机号")
    email = models.EmailField(blank=True, null=True, verbose_name="邮箱")
    avatar = models.TextField(blank=True, null=True, verbose_name='头像')
    route = models.JSONField(blank=True, null=True, verbose_name='后端接口权限')
    department = models.ForeignKey("Department", blank=True, null=True, default='87654321876543218765432187654321',
                                   related_name='departments', on_delete=models.CASCADE, db_constraint=False,
                                   verbose_name='所属部门')
    REQUIRED_FIELDS = []  # `python manage.py createsuperuser`执行创建超级用户时，无需填写邮箱

    class Meta:
        db_table = 'auth_user'
        verbose_name = '用户表'
        verbose_name_plural = verbose_name


class Role(BaseModel):
    name = models.CharField(max_length=64, blank=True, null=True, verbose_name="角色名称")
    sort = models.SmallIntegerField(default=1, verbose_name='sort')
    is_active = models.BooleanField(default=True, verbose_name="是否启用")
    remark = models.CharField(max_length=128, blank=True, null=True, verbose_name="备注")
    menu_ids = MultipleChoiceField(blank=True, null=True, verbose_name='菜单集合')
    users = models.ManyToManyField('User', blank=True, null=True, verbose_name='关联用户', db_constraint=False,
                                   related_name='roles')

    class Meta:
        db_table = 'system_role'
        verbose_name = '角色表'
        verbose_name_plural = verbose_name


class Menu(BaseModel):
    parent = FixCharField(max_length=32, default='12345678901234567890123456789012', verbose_name='上级目录')
    name = models.CharField(max_length=64, blank=True, null=True, verbose_name='名称')
    category = models.CharField(max_length=64, blank=True, null=True, verbose_name='类型')
    path = models.CharField(max_length=64, blank=True, null=True, verbose_name='路由路径')
    component = models.CharField(max_length=64, blank=True, null=True, verbose_name='组件')
    sort = models.SmallIntegerField(default=1, verbose_name='排序')
    icon = models.CharField(max_length=64, blank=True, null=True, verbose_name='图标')
    redirect = models.CharField(max_length=64, blank=True, null=True, verbose_name='重定向地址')
    perm = models.CharField(max_length=32, blank=True, null=True, verbose_name='按钮权限')
    hidden = models.BooleanField(default=False, verbose_name='是否隐藏')
    keep_alive = models.BooleanField(default=True, verbose_name='是否缓存')
    always_show = models.BooleanField(default=False, verbose_name="根节点是否始终显示")

    class Meta:
        db_table = 'system_menu'
        verbose_name = '菜单表'
        verbose_name_plural = verbose_name


class Attachment(BaseModel):
    name = models.CharField(max_length=64, blank=True, null=True, verbose_name="附件名称")
    file = models.FileField(upload_to='%Y-%m-%d/', verbose_name="file")
    username = models.CharField(max_length=150, blank=True, null=True, verbose_name="用户名")

    class Meta:
        db_table = 'system_attachment'
        verbose_name = '附件表'
        verbose_name_plural = verbose_name


@receiver(pre_save, sender=Attachment)
def rename_file(sender, instance, **kwargs):
    """
    上传模型中file字段，重命名存储，去掉特殊符号，避免安全问题
    """
    filename = instance.file.name
    extension = filename.split('.')[-1]
    name = '.'.join(filename.split('.')[:-1])
    instance.file.name = slugify(name, allow_unicode=True) + '.' + extension


class Lookup(BaseModel):
    category = models.CharField(max_length=32, verbose_name="类型")
    name = models.CharField(max_length=32, blank=True, null=True, verbose_name="类型名称")
    value = models.CharField(max_length=32, blank=True, null=True, verbose_name="值")
    label = models.CharField(max_length=64, blank=True, null=True, verbose_name="值的含义")
    sort = models.SmallIntegerField(default=1, verbose_name='排序')
    is_active = models.BooleanField(default=True, verbose_name="是否启用")
    is_category = models.BooleanField(default=True, verbose_name='是否是类型')
    remark = models.CharField(max_length=128, blank=True, null=True, verbose_name="备注")

    class Meta:
        db_table = 'system_lookup'
        verbose_name = '码表'
        verbose_name_plural = verbose_name


@cache()
def get_lookup():
    result = {}
    for lookup in Lookup.objects.all():
        if result.get(lookup.category):
            result[lookup.category].append(lookup)
        else:
            result[lookup.category] = [lookup]
    return result


def get_value_by_category(category):
    data = get_lookup()
    res = []
    if data.get(category):
        res = [o.value for o in data[category] if o.is_active]
    return res
