import logging

from django.http import FileResponse
from rest_framework.decorators import action

from app.settings import DEFAULT_USER_PASSWORD
from system.models import User, Role, Menu, Attachment, Lookup, Department
from system.serializers import UserSerializer, RoleSerializer, MenuSerializer, AttachmentSerializer, LookupSerializer, \
    DepartmentSerializer, UserListSerializer
from utils.response import SuccessResponse
from utils.util import build_tree
from utils.viewset import BaseModelViewSet, BaseViewSet

logger = logging.getLogger()


class DepartmentViewSet(BaseModelViewSet):
    queryset = Department.objects.all()
    serializer_class = DepartmentSerializer
    route_label = '部门模块'
    filterset_fields = {"name": ["icontains"], "is_active": ["exact"]}

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        ids = [o.id for o in queryset]
        sql = f"""
WITH RECURSIVE hierarchy AS (
    SELECT * FROM system_department where id in %s
    UNION ALL
    SELECT t.* FROM system_department t
    JOIN hierarchy h ON h.parent = t.id )
SELECT DISTINCT * FROM hierarchy ORDER BY sort;
"""
        # WITH RECURSIVE是mysql的内置的递归查询方法，从下往上查询
        # 如需从上往下查询，只需修改 JOIN hierarchy h ON t.parent = h.id
        data = DepartmentSerializer(Department.objects.raw(sql, [ids]), many=True).data
        return SuccessResponse(build_tree(data))


class UserViewSet(BaseModelViewSet):
    queryset = User.objects.order_by('-create_time')
    serializer_class = UserSerializer
    serializer_list_class = UserListSerializer
    route_label = '用户模块'
    route_method_extend = {'reset_password': '重置用户密码'}
    filterset_fields = {"username": ["icontains"], "nickname": ["icontains"], "is_active": ["exact"],
                        "department": ["exact"]}

    def create(self, request, *args, **kwargs):
        # 创建用户时使用默认密码
        response = super().create(request, *args, **kwargs)
        user = (User.objects.get(id=response.data['data']['id']))
        user.set_password(DEFAULT_USER_PASSWORD)
        user.save(update_fields=["password"])
        return response

    @action(methods=['post'], detail=False)
    def reset_password(self, request):
        '''
        重置默认密码,提交需要重置用户的id，参数 ['id', 'id',...]
        '''
        for id in request.data:
            user = User.objects.get(id=id)
            user.set_password(DEFAULT_USER_PASSWORD)
            user.save(update_fields=["password"])
        return SuccessResponse()


class RoleViewSet(BaseModelViewSet):
    queryset = Role.objects.order_by('sort')
    serializer_class = RoleSerializer
    route_label = '角色模块'
    filterset_fields = {"name": ["icontains"], "is_active": ["exact"]}


class MenuViewSet(BaseModelViewSet):
    queryset = Menu.objects.all()
    serializer_class = MenuSerializer
    route_label = '菜单模块'
    filterset_fields = {"name": ["icontains"], "category": ["exact"], "perm": ["icontains"], "hidden": ["exact"]}

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        ids = [o.id for o in queryset]
        sql = f"""
WITH RECURSIVE hierarchy AS (
    SELECT * FROM system_menu where id in %s
    UNION ALL
    SELECT t.* FROM system_menu t
    JOIN hierarchy h ON h.parent = t.id )
SELECT DISTINCT * FROM hierarchy ORDER BY sort;
"""
        data = MenuSerializer(Menu.objects.raw(sql, [ids]), many=True).data
        return SuccessResponse(build_tree(data))


class AttachmentViewSet(BaseModelViewSet):
    queryset = Attachment.objects.all()
    serializer_class = AttachmentSerializer
    route_label = '附件模块'
    route_method_extend = {'download': '附件下载'}
    route_method_exclude = ['update']

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        instance.file.delete()
        self.perform_destroy(instance)
        return SuccessResponse()

    @action(methods=['post'], detail=False)
    def batch_destroy(self, request):
        queryset = self.get_queryset()
        objects = queryset.filter(id__in=request.data).all()
        for o in objects:
            o.file.delete()
            o.delete()
        return SuccessResponse()

    @action(methods=['get'], detail=False, url_path=r'download/(?P<pk>[^/.]+)')
    def download(self, request, pk):
        instance = self.get_object()
        return FileResponse(instance.file.file)


class LookupViewSet(BaseModelViewSet):
    queryset = Lookup.objects.order_by('sort', 'name')
    serializer_class = LookupSerializer
    route_label = '码表模块'
    filterset_fields = {"category": ["exact"], "name": ["icontains"], "is_active": ["exact"]}

    def list(self, request, *args, **kwargs):
        lookups = {}  # 预处理数据成dict，键为category的值
        for lookup in Lookup.objects.filter(is_category=False).order_by('sort'):
            data = self.get_list_serializer(lookup).data
            if lookups.get(lookup.category):
                lookups[lookup.category].append(data)
            else:
                lookups[lookup.category] = [data]

        queryset = Lookup.objects.filter(is_category=True)
        queryset = self.filter_queryset(queryset)
        page = self.paginate_queryset(queryset)
        res = []
        if page is not None:
            for row in page:
                data = LookupSerializer(row).data
                data['children'] = lookups.get(row.category)
                res.append(data)
            return SuccessResponse(self.get_paginated_response(res))
        for row in queryset:
            data = LookupSerializer(row).data
            data['children'] = lookups.get(row.category)
            res.append(data)
        return SuccessResponse(res)


class ExtViewSet(BaseViewSet):
    route_label = '系统拓展模块'
    route_method = {'get_route_template': '获取后端权限模版'}
    route_method_allow = {
        'get_own_info': '获取用户信息',
        'set_own_info': '修改用户信息',
        'get_own_menu': '获取菜单',
        'set_own_password': '修改密码',
        'get_own_lookup': '获取码表',
        'logout': '登出'
    }

    def _get_user_menu_ids(self, user):
        # 根据user获取可用角色id和角色上获取menu的id集合
        menu_ids = set()
        role_ids = []
        for role in user.roles.filter(is_active=True):
            role_ids.append(role.id)
            if not role.menu_ids:
                continue
            menu_ids.update(role.menu_ids)
        return list(menu_ids), role_ids

    @action(methods=['get'], detail=False)
    def get_own_info(self, request):
        """
        用户登录后获取所需的信息
        """
        data = UserSerializer(request.user).data
        menu_ids, role_ids = self._get_user_menu_ids(request.user)
        data['roles'] = role_ids
        if request.user.is_superuser:
            data['perms'] = [menu.perm for menu in Menu.objects.all() if menu.perm]
        else:
            data['perms'] = [menu.perm for menu in Menu.objects.filter(id__in=menu_ids) if menu.perm]
        return SuccessResponse(data)

    @action(methods=['post'], detail=False)
    def set_own_info(self, request):
        """
        修改个人信息，参数：{"nickname": "xx", "phone": "xx", "email": "xx", "avatar": "xx"}
        """
        request.user.nickname = request.data.get('nickname')
        request.user.phone = request.data.get('phone')
        request.user.email = request.data.get('email')
        request.user.avatar = request.data.get('avatar')
        request.user.save(update_fields=["nickname", "phone", "email", "avatar"])
        return SuccessResponse()

    @action(methods=['get'], detail=False)
    def get_own_menu(self, request):
        """
        用户登录后获取所需的菜单
        """
        queryset = Menu.objects.order_by('sort')
        # 如果是超级用户则获取全部menu树结构
        if request.user.is_superuser:
            tree = build_tree(MenuSerializer(queryset, many=True).data)
            return SuccessResponse(tree)
        # 普通用户按照角色查找menu树结构
        menu_ids, _ = self._get_user_menu_ids(request.user)
        tree = build_tree(MenuSerializer(queryset.filter(id__in=menu_ids), many=True).data)
        return SuccessResponse(tree)

    @action(methods=['get'], detail=False)
    def get_own_lookup(self, request):
        """
        用户登录后获取码表
        """
        lookups = Lookup.objects.filter(is_active=True, is_category=False).order_by('sort')
        res = {}
        for lookup in lookups:
            item = {'label': lookup.label, 'value': lookup.value}
            if res.get(lookup.category):
                res[lookup.category].append(item)
            else:
                res[lookup.category] = [item]
        return SuccessResponse(res)

    @action(methods=['post'], detail=False)
    def set_own_password(self, request):
        """
        修改登录用户的密码,参数 {"password": "xxx", "old_password": "xxx"}
        """
        if not request.user.check_password(request.data['old_password']):
            raise Exception("旧密码输入错误")
        request.user.set_password(request.data['password'])
        request.user.save(update_fields=["password"])
        return SuccessResponse()

    @action(methods=['get'], detail=False)
    def get_route_template(self, request):
        """
        后端权限模版
        """
        from app.urls import router
        result = []
        for obj in router.registry:
            route_label, route_method = obj[1].get_route()
            row = {'route_name': obj[1].__name__, 'route_label': route_label, 'route_method': route_method}
            result.append(row)
        return SuccessResponse(result)
