from collections import OrderedDict

from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets, mixins
from rest_framework.decorators import action
from rest_framework.filters import SearchFilter, OrderingFilter
from rest_framework.pagination import PageNumberPagination
from rest_framework.viewsets import GenericViewSet

from utils.execption import CustomServException
from utils.permission import RoutePermission, IsAuthPermission
from utils.response import SuccessResponse


class Pagination(PageNumberPagination):
    page_size_query_param = 'size'
    page_query_param = 'current'

    def get_paginated_response(self, data):
        return OrderedDict([
            ('total', self.page.paginator.count),
            ('results', data)
        ])


class BaseGenericViewSet(GenericViewSet):
    permission_classes = (IsAuthPermission, RoutePermission)

    @classmethod
    def get_route(cls):
        if not hasattr(cls, 'route_label'):
            raise CustomServException("未完善route_label属性")
        route_method = {}
        if issubclass(cls, BaseListModelMixin):
            route_method.update({'list': '查询'})
        if issubclass(cls, BaseRetrieveModelMixin):
            route_method.update({'retrieve': '详细信息'})
        if issubclass(cls, BaseCreateModelMixin):
            route_method.update({'create': '新增'})
        if issubclass(cls, BaseUpdateModelMixin):
            route_method.update({'update': '修改', 'partial_update': '部分修改', 'batch_update': '批量修改'})
        if issubclass(cls, BaseDestroyModelMixin):
            route_method.update({'destroy': '删除', 'batch_destroy': '批量删除'})
        if hasattr(cls, 'route_method'):
            route_method = cls.route_method
        if hasattr(cls, 'route_method_extend'):
            route_method.update(cls.route_method_extend)
        if hasattr(cls, 'route_method_exclude'):
            for k in cls.route_method_exclude:
                route_method.pop(k)
        return cls.route_label, route_method


class BaseListModelMixin(mixins.ListModelMixin, BaseGenericViewSet):
    pagination_class = Pagination
    filter_backends = (DjangoFilterBackend, SearchFilter, OrderingFilter)

    def paginate_queryset(self, queryset):
        if self.request.GET.get('size') and self.request.GET.get('current'):
            return super().paginate_queryset(queryset)
        return None

    def get_list_serializer(self, *args, **kwargs):
        """
        Return the serializer instance that should be used for validating and
        deserializing input, and for serializing output.
        """
        if hasattr(self, 'serializer_list_class'):
            serializer_class = self.serializer_list_class
        else:
            serializer_class = self.get_serializer_class()
        kwargs.setdefault('context', self.get_serializer_context())
        return serializer_class(*args, **kwargs)

    def list(self, request, *args, **kwargs):
        """
        size: 分页，展示条数
        current: 分页，当前第N页，size和current同时存在开启分页查询
        search: 模糊搜索，后端配合添加属性 search_fields=["name", "remark"]
        ordering: 排序，后端需要配置添加属性 ordering_fields=["create_time", "-update_time"]
        高级查询:
            filterset_fields精准查询，按需配置filterset_fields=["name", "phone"]  # filterset_fields也如下配置，icontains、exact等关键字查看README
            filterset_fields = {"username": ["icontains"], "nickname": ["icontains"], "is_active": ["exact"], "department": ["exact"]}
            filterset_class自定义高级查询，按需完善filterset_class配置类
        """
        queryset = self.filter_queryset(self.get_queryset())
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_list_serializer(page, many=True)
            return SuccessResponse(self.get_paginated_response(serializer.data))
        serializer = self.get_list_serializer(queryset, many=True)
        return SuccessResponse(serializer.data)


class BaseRetrieveModelMixin(mixins.RetrieveModelMixin, BaseGenericViewSet):
    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        return SuccessResponse(serializer.data)


class BaseCreateModelMixin(mixins.CreateModelMixin, BaseGenericViewSet):
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        return SuccessResponse(serializer.data)


class BaseUpdateModelMixin(mixins.UpdateModelMixin, BaseGenericViewSet):
    def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(instance, '_prefetched_objects_cache', None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return SuccessResponse(serializer.data)

    def partial_update(self, request, *args, **kwargs):
        kwargs['partial'] = True
        return self.update(request, *args, **kwargs)

    @action(methods=['post'], detail=False)
    def batch_update(self, request):
        """
        参数：{"selected_ids": ["id", "id",...], "update_field": {"is_active": false, "name": "xxx"}}
        """
        queryset = self.get_queryset()
        queryset.filter(id__in=request.data['selected_ids']).update(**request.data['update_field'])
        return SuccessResponse()


class BaseDestroyModelMixin(mixins.DestroyModelMixin, BaseGenericViewSet):
    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        self.perform_destroy(instance)
        return SuccessResponse()

    @action(methods=['post'], detail=False)
    def batch_destroy(self, request):
        """
        参数：["id", "id",...]
        """
        queryset = self.get_queryset()
        queryset.filter(id__in=request.data).delete()
        return SuccessResponse()


class BaseModelViewSet(BaseListModelMixin,
                       BaseRetrieveModelMixin,
                       BaseCreateModelMixin,
                       BaseUpdateModelMixin,
                       BaseDestroyModelMixin):
    """
    queryset = User.objects.all()
    serializer_class = UserSerializer
    serializer_list_class = UserListSerializer  # 列表查询所需的序列化，不写默认使用serializer_class序列化
    route_label = 'xx模块'
    route_method = {'method_name': '该方法中文简短描述'}  # 权限方法
    route_method_extend = {'method_name': '该方法中文简短描述'}  # 权限拓展
    route_method_exclude = ['method_name']  # 权限排除
    route_method_allow = {'method_name': '该方法中文简短描述'}  # 权限一直放行
    """
    pass


class BaseViewSet(viewsets.ViewSet):
    '''
    route_label = 'xx模块'
    route_method = {'method_name': '该方法中文简短描述'}  # 权限方法
    route_method_allow = {'method_name': '该方法中文简短描述'}  # 权限一直放行
    '''
    permission_classes = (IsAuthPermission, RoutePermission)

    @classmethod
    def get_route(cls):
        if not hasattr(cls, 'route_label'):
            raise CustomServException("未完善route_label属性")
        route_method = None
        if hasattr(cls, 'route_method'):
            route_method = cls.route_method
        return cls.route_label, route_method
