import base64
import json
import shutil
from concurrent.futures.thread import ThreadPoolExecutor

from rest_framework.decorators import action

from asset.models import Agent, TaskAgent, TaskAgentData
from migrate_check import check_account, check_repo, check_gz
from migrate_check_ext import get_info, get_local, get_config, get_remote, get_centos_version, get_centos_core_list, \
    check_abi
from thread_queue import update_status
from utils.response import SuccessResponse, ErrorResponse
from utils.util import RemoteExec
from utils.viewset import BaseViewSet


class ValidationViewSet(BaseViewSet):
    route_label = '检测模块'
    route_method = {'get_info_by_taskagent': '获取机器上的信息',
                    'check_account_by_task': '检测agent账号是否可以登录并是否有root权限',
                    'check_repo_by_task': '检测agent上的repo是否可用',
                    'check_data_by_task': '迁移前检查数据缓存',
                    'get_data_by_taskagent': '迁移前检查',
                    'check_abi_by_task': 'abi检测',
                    'get_abi_by_taskagent': '获取abi检测数据',
                    'get_data_migrate_by_taskagent': '迁移后检查',
                    'check_gz_by_task': '检测agent上的迁移文件'}

    @action(methods=['get'], detail=False)
    def get_info_by_taskagent(self, request):
        """
        通过taskagent获取agent机器上的信息,参数 get_info_by_taskagent/?taskagent_id=xxx
        """
        taskagent_id = request.query_params.get('taskagent_id')
        agent = Agent.objects.filter(taskagents__id=taskagent_id).first()
        if agent is None:
            raise Exception("agent_id不存在")
        remote = (agent.ip, agent.port, agent.username, agent.password)
        try:
            data = get_info(remote)
        except Exception:
            raise Exception(f'{agent.ip}无法连接')
        return SuccessResponse(data=data)

    @action(methods=['get'], detail=False)
    def check_account_by_task(self, request):
        """
        通过task_id检测agent账号是否可以登录并是否有root权限,参数 check_account_by_task/?task_id=xxx
        """
        task_id = request.query_params.get('task_id')
        taskagents = TaskAgent.objects.filter(task__id=task_id)
        pool = ThreadPoolExecutor(max_workers=len(taskagents))
        items = []
        for taskagent in taskagents:
            if taskagent.task.plan == "A":
                update_status(taskagent, 1, '')
            elif taskagent.task.plan == "B":
                update_status(taskagent, 6, '')
            agent = taskagent.agent
            remote = (agent.ip, agent.port, agent.username, agent.password)
            items.append((taskagent, pool.submit(check_account, remote)))
        pool.shutdown()
        res = []
        for item in items:
            taskagent, p = item
            is_success, msg = p.result()
            if taskagent.task.plan == "A":
                update_status(taskagent, 10 if is_success else 100, msg)
            elif taskagent.task.plan == "B":
                update_status(taskagent, 60 if is_success else 600, msg)
            res.append({"taskagent_id": taskagent.id, "is_success": is_success, "msg": msg})
        return SuccessResponse(res)

    @action(methods=['get'], detail=False)
    def check_repo_by_task(self, request):
        """
        通过task_id检测agent上的repo是否可用，参数 check_repo_by_task/?task_id=xxx
        """
        task_id = request.query_params.get('task_id')
        taskagents = TaskAgent.objects.filter(task__id=task_id)
        pool = ThreadPoolExecutor(max_workers=len(taskagents))
        items = []
        for taskagent in taskagents:
            update_status(taskagent, 2, '')
            agent = taskagent.agent
            remote = (agent.ip, agent.port, agent.username, agent.password)
            repo_content = taskagent.task.repo.content
            if taskagent.task.repo and taskagent.task.repo.content:
                items.append((taskagent, pool.submit(check_repo, remote, repo_content)))
            else:
                return ErrorResponse(msg='请检查迁移源是否有完善')
        pool.shutdown()
        res = []
        for item in items:
            taskagent, p = item
            is_success, msg = p.result()
            update_status(taskagent, 20 if is_success else 200, msg)
            res.append({"taskagent_id": taskagent.id, "is_success": is_success, "msg": msg})
        return SuccessResponse(res)

    @action(methods=['get'], detail=False)
    def check_data_by_task(self, request):
        """
        通过task_id保存agent上的rpm,系统信息，配置信息,参数 check_data_by_task/?task_id=xxx
        """
        task_id = request.query_params.get('task_id')
        taskagents = TaskAgent.objects.filter(task__id=task_id)
        pool = ThreadPoolExecutor(max_workers=len(taskagents))
        for taskagent in taskagents:
            update_status(taskagent, 3, '')
            agent = taskagent.agent
            remote = (agent.ip, agent.port, agent.username, agent.password)
            pool.submit(check_data, remote, taskagent)
        return SuccessResponse()

    @action(methods=['get'], detail=False)
    def get_data_by_taskagent(self, request):
        """
        通过taskagent获取agent检测信息,参数 get_data_by_taskagent/?taskagent_id=xxx
        """
        taskagent_id = request.query_params.get('taskagent_id')
        taskagent = TaskAgent.objects.get(id=taskagent_id)
        data = taskagent.taskagentdata
        if data.info_old and data.rpm_old and data.config_old:
            agent = taskagent.agent
            remote = (agent.ip, agent.port, agent.username, agent.password)
            repo_content = taskagent.task.repo.content
            remotes = get_remote(remote, repo_content)
            res = {'config': json.loads(data.config_old)}  # 配置
            rpm_old = json.loads(data.rpm_old)
            res['old_package_num'] = len(rpm_old.keys())  # 原系统包数量
            res['nfs_package_num'] = len(remotes.keys())  # 方德包数量
            res['replace_package_num'] = len(set(rpm_old.keys()) & set(remotes.keys()))  # 待替换包数量
            res['lost_package_num'] = len(set(rpm_old.keys()) - set(remotes.keys()))  # 问题包数量
            res['replace_packages'] = [rpm_old[k] for k in set(rpm_old.keys()) & set(remotes.keys())]  # 待替换包
            res['lost_packages'] = [rpm_old[k] for k in set(rpm_old.keys()) - set(remotes.keys())]  # 缺失包
            # 高风险项包含在方德操作系统中不存在的属于最小化安装范围的包、i686的安装包     有一个包在原系统是属于minimal的组，但是方德os里面没有的
            centos_version = get_centos_version(remote)
            centos_core_list = get_centos_core_list(centos_version)
            res['high_risk_packages'] = [rpm_old[k] for k in set(centos_core_list) - set(remotes.keys()) if rpm_old.get(k)]
            return SuccessResponse(data=res)
        else:
            return SuccessResponse(msg='检查中，请稍后。')

    @action(methods=['get'], detail=False)
    def check_abi_by_task(self, request):
        """
        通过task_id检测agent上的abi对比信息,参数 check_abi_by_task/?task_id=xxx
        """
        task_id = request.query_params.get('task_id')
        taskagents = TaskAgent.objects.filter(task__id=task_id)
        if all([int(taskagent.status) in [4, 40, 400] for taskagent in taskagents]):
            return SuccessResponse(msg='检测执行中或已经完成')
        pool = ThreadPoolExecutor(max_workers=len(taskagents))
        items = []
        for taskagent in taskagents:
            agent = taskagent.agent
            remote = (agent.ip, agent.port, agent.username, agent.password)
            items.append((taskagent, pool.submit(check_abi, remote)))
        pool.shutdown()
        res = []
        for item in items:
            taskagent, p = item
            is_success, msg = p.result()
            res.append({"taskagent_id": taskagent.id, "is_success": is_success, "msg": msg, "ip": taskagent.agent.ip})

        if all([i["is_success"] for i in res]):
            # 所有都为True则运行abi
            pool1 = ThreadPoolExecutor(max_workers=len(taskagents))
            for taskagent in taskagents:
                update_status(taskagent, 4, '')
                agent = taskagent.agent
                remote = (agent.ip, agent.port, agent.username, agent.password)
                pool1.submit(run_abi, remote, taskagent)
        return SuccessResponse(res)

    @action(methods=['get'], detail=False)
    def get_abi_by_taskagent(self, request):
        """
        通过taskagent获取agent上的abi对比信息,参数 get_abi_by_taskagent/?taskagent_id=xxx
        """
        taskagent_id = request.query_params.get('taskagent_id')
        taskagent = TaskAgent.objects.get(id=taskagent_id)
        agent = taskagent.agent
        res = {}
        with RemoteExec(ip=agent.ip, port=agent.port, username=agent.username, password=agent.password) as ssh:
            res['os_abi'] = ssh.exec('cat /home/nfs_migration_abi/os_abi/out/os_abi/os_abi_compare.nadt')
            res['os_kabi'] = ssh.exec('cat /home/nfs_migration_abi/os_kabi/out/os_kabi/os_kabi_compare.nadt')
            res['os_kconfig'] = ssh.exec('cat /home/nfs_migration_abi/os_kconfig/out/os_kconfig/os_kconfig_compare.nadt')
            res['os_rpm_depends'] = ssh.exec('cat /home/nfs_migration_abi/os_rpm_depends/out/osrpmdepends/osrpmdepends_compare.nadt')
            res['os_rule'] = ssh.exec('cat /home/nfs_migration_abi/os_rule/out/osruler/osruler_compare.nadt')
            res['os_service_config'] = ssh.exec('cat /home/nfs_migration_abi/os_service_config/out/osservice-config/osservice-config_compare.nadt')
            res['os_system'] = ssh.exec('cat /home/nfs_migration_abi/os_system/out/ossystem/ossystem_compare.nadt')
            try:
                res['os_abi'] = json.loads(base64.b64decode(res['os_abi']).decode('utf-8'))
                res['os_kabi'] = json.loads(base64.b64decode(res['os_kabi']).decode('utf-8'))
                res['os_kconfig'] = json.loads(base64.b64decode(res['os_kconfig']).decode('utf-8'))
                res['os_rpm_depends'] = json.loads(base64.b64decode(res['os_rpm_depends']).decode('utf-8'))
                res['os_rule'] = json.loads(base64.b64decode(res['os_rule']).decode('utf-8'))
                res['os_service_config'] = json.loads(base64.b64decode(res['os_service_config']).decode('utf-8'))
                res['os_system'] = json.loads(base64.b64decode(res['os_system']).decode('utf-8'))
            except:
                if int(taskagent.status) == 4:
                    return SuccessResponse(msg='检测中，请稍后。')
                else:
                    return SuccessResponse(msg='检测未开始。')
        return SuccessResponse(res)

    @action(methods=['get'], detail=False)
    def get_data_migrate_by_taskagent(self, request):
        """
        通过taskagent获取agent迁移后信息,参数 get_data_migrate_by_taskagent/?taskagent_id=xxx
        """
        taskagent_id = request.query_params.get('taskagent_id')
        data = TaskAgentData.objects.filter(task_agent_id=taskagent_id).first()
        if data.info_new and data.rpm_new and data.config_new:
            res = {}
            res['info_old'] = json.loads(data.info_old)
            res['info_new'] = json.loads(data.info_new)
            res['rpm_old'] = json.loads(data.rpm_old)
            res['rpm_new'] = json.loads(data.rpm_new)
            res['config_old'] = json.loads(data.config_old)
            res['config_new'] = json.loads(data.config_new)
            return SuccessResponse(data=res)
        else:
            return SuccessResponse(msg='检查中，请稍后。')

    @action(methods=['get'], detail=False)
    def check_gz_by_task(self, request):
        """
        通过task_id检测agent上的repo是否可用，不可用则还原repo,参数 check_gz_by_task/?task_id=xxx&path_gz=/home/os_migrate.tar.gz
        """
        task_id = request.query_params.get('task_id')
        path_gz = request.query_params.get('path_gz')
        taskagents = TaskAgent.objects.filter(task__id=task_id)
        pool = ThreadPoolExecutor(max_workers=len(taskagents))
        items = []
        for taskagent in taskagents:
            update_status(taskagent, 7, '')
            agent = taskagent.agent
            remote = (agent.ip, agent.port, agent.username, agent.password)
            items.append((taskagent, pool.submit(check_gz, remote, path_gz)))
            taskagent.task.path_gz = path_gz
            taskagent.task.save(update_fields=["path_gz"])
        pool.shutdown()
        res = []
        for item in items:
            taskagent, p = item
            is_success, msg = p.result()
            update_status(taskagent, 70 if is_success else 700, msg)
            res.append({"taskagent_id": taskagent.id, "is_success": is_success, "msg": msg})
        return SuccessResponse(res)


def check_data(remote, taskagent):
    task_agent_data = taskagent.taskagentdata
    task_agent_data.info_old = json.dumps(get_info(remote))
    task_agent_data.rpm_old = json.dumps(get_local(remote))
    task_agent_data.config_old = json.dumps(get_config(remote))
    task_agent_data.save()
    update_status(taskagent, 30, '成功')


def run_abi(remote, taskagent):
    ip, port, username, password = remote
    with RemoteExec(ip=ip, port=port, username=username, password=password) as ssh:
        sftp = ssh.get_sftp()
        with sftp.open('/home/nfs_migration_abi.tar.gz', 'wb') as fp:
            shutil.copyfileobj(open('/home/nfs_migration_abi.tar.gz', 'rb'), fp)
        # sftp.put('/home/nfs_migration_abi.tar.gz', '/home/nfs_migration_abi.tar.gz')
        ssh.exec('cd /home && tar -xzvf nfs_migration_abi.tar.gz')
        if ssh.exec('rpm -qa | grep libabigail') == '':
            ssh.exec('cd /home/nfs_migration_abi/rpm && rpm -ivh libabigail-1.6-4.x86_64.rpm --nodeps')
        if get_centos_version(remote) in ['7.0', '7.1', '7.2', '7.3', '7.4', '7.5', '7.6']:
            if ssh.exec('rpm -qa | grep python3-rpm') == '':
                ssh.exec('cd /home/nfs_migration_abi/rpm && rpm -ivh python3-rpm-4.11.3-7.nfs.x86_64.rpm --nodeps')

        os_abi = ssh.exec('bash /home/nfs_migration_abi/os_abi/run.sh')
        ssh.file_a('/home/nfs_migration_abi/os_abi/out/run.log', os_abi)

        os_kabi = ssh.exec('bash /home/nfs_migration_abi/os_kabi/run.sh')
        ssh.file_a('/home/nfs_migration_abi/os_kabi/out/run.log', os_kabi)

        os_kconfig = ssh.exec('bash /home/nfs_migration_abi/os_kconfig/run.sh')
        ssh.file_a('/home/nfs_migration_abi/os_kconfig/out/run.log', os_kconfig)

        os_rpm_depends = ssh.exec('bash /home/nfs_migration_abi/os_rpm_depends/run.sh')
        ssh.file_a('/home/nfs_migration_abi/os_rpm_depends/out/run.log', os_rpm_depends)

        os_rule = ssh.exec('bash /home/nfs_migration_abi/os_rule/run.sh')
        ssh.file_a('/home/nfs_migration_abi/os_rule/out/run.log', os_rule)

        os_service_config = ssh.exec('bash /home/nfs_migration_abi/os_service_config/run.sh')
        ssh.file_a('/home/nfs_migration_abi/os_service_config/out/run.log', os_service_config)

        os_system = ssh.exec('bash /home/nfs_migration_abi/os_system/run.sh')
        ssh.file_a('/home/nfs_migration_abi/os_system/out/run.log', os_system)

    update_status(taskagent, 40, '成功')
